/*
 * ==================================================================
 *
 * prelink_tool: simple utility for processing (decompressing or
 * recompressing) the AppleTV's "mach_kernel.prelink" file.
 *
 * Written 2007/03/29 by Paul Bartholomew (oz_paulb@hotmail.com)
 *
 * Thanks to info at http://0xfeedbeef.com/appletv/ for pointer to
 * Apple's(?) "lzwss" compression/decompression source code.
 *
 * ==================================================================
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "mytypes.h"
#include "compression.h"
#include "utils.h"


int	decode_lzss_prelink_file(unsigned char *p_buf, unsigned int sz_buf, const char *out_fname);
int	encode_lzss_prelink_file(unsigned char *p_buf, unsigned int sz_buf, const char *out_fname);
void	usage(void);

const char	*Pname;

int
main(int argc, char *argv[])
{
	int	want_decode;
	int	narg, argx;
	const char	*in_fname, *out_fname;
	unsigned char	*p_infile;
	unsigned int	sz_infile;


	printf("================================================\n"); fflush(stdout);
	printf("prelink_tool: mach_kernel.prelink file processor\n\tby Paul Bartholomew (oz_paulb)\n"); fflush(stdout);
	printf("================================================\n\n"); fflush(stdout);

	want_decode = 1;

	Pname = argv[0];

	argx = 1;
	narg = argc-1;
	while(narg) {
		if (argv[argx][0] == '-') {
			if (!strcmp(argv[argx], "-d")) {
				want_decode = 1;
			} else if (!strcmp(argv[argx], "-e")) {
				want_decode = 0;
			}
		} else {
			break;
		}
		++argx;
		--narg;
	}

	if (narg != 2) {
		usage();
		exit(1);
	}

	in_fname = argv[argx++];
	out_fname = argv[argx++];

//	printf("%s: in_fname = %s, out_fname = %s\n", want_decode ? "decode" : "encode", in_fname, out_fname);

	if (!load_binary_file(in_fname, &p_infile, &sz_infile)) {
		fprintf(stderr, "Error loading input file: '%s'!\n", in_fname);
		exit(1);
	}

	printf("Loaded input file: '%s', size = %u\n", in_fname, sz_infile); fflush(stdout);

	if (want_decode) {
		decode_lzss_prelink_file(p_infile, sz_infile, out_fname);
	} else {
		encode_lzss_prelink_file(p_infile, sz_infile, out_fname);
	}
	free((void *)p_infile);

	return 0;
}

#define	DECOMP_BUFFER_PAD		65536

int
decode_lzss_prelink_file(unsigned char *p_buf, unsigned int sz_buf, const char *out_fname)
{
	PRELINK_LZSS_FILE_HEADER	fhdr;
	unsigned char	*p, *p_payload;
	unsigned int	sz_payload;
	int	decomp_ret;
	unsigned int	out_file_size;
	unsigned char	*p_decomp_buf;
	unsigned int	decomp_buffer_size;
	unsigned int	computed_adler32;


	memset(&fhdr, 0, sizeof(fhdr));

	//
	// read file header into local structure, converting
	// from file's "big endian" format to host format 
	// along the way.
	//
	p = p_buf;
	p = read_mem_BE_U32(p, &fhdr.signature);
	p = read_mem_BE_U32(p, &fhdr.compress_type);
	p = read_mem_BE_U32(p, &fhdr.adler32);
	p = read_mem_BE_U32(p, &fhdr.uncompressed_size);
	p = read_mem_BE_U32(p, &fhdr.compressed_size);
	p = read_mem_bytes(p, &fhdr.reserved_1, sizeof(fhdr.reserved_1));
	p = read_mem_bytes(p, &fhdr.reserved_2, sizeof(fhdr.reserved_2));
	p = read_mem_bytes(p, &fhdr.filename, sizeof(fhdr.filename));


	p_payload = (p_buf + PRELINK_LZSS_FILE_HEADER_SIZE_ON_DISK);
	sz_payload = (sz_buf - PRELINK_LZSS_FILE_HEADER_SIZE_ON_DISK);

#if	0
	printf("sig = 0x%x, compr = 0x%x\n", fhdr.signature, fhdr.compress_type);
	printf("adler32 = 0x%x\n", fhdr.adler32);
	printf("uncomp_size = 0x%x (%u)\n", fhdr.uncompressed_size, fhdr.uncompressed_size);
	printf("comp_size = 0x%x (%u)\n", fhdr.compressed_size, fhdr.compressed_size);

	printf("computed_adler32 (compr) = 0x%x\n", local_adler32(p_payload, sz_payload));
	printf("filename = '%s'\n", fhdr.filename);
#endif

	// do some sanity checking on the file header

	if ((fhdr.signature != PRELINK_LZSS_FILE_HEADER__SIGNATURE) || 
		(fhdr.compress_type != PRELINK_LZSS_FILE_HEADER__COMPRESS_TYPE)) {
		fprintf(stderr, "ERROR: Header signature/compression type mismatch!\n");
		return 0;
	}


	if ((fhdr.compressed_size + PRELINK_LZSS_FILE_HEADER_SIZE_ON_DISK) != sz_buf) {
		fprintf(stderr, "WARNING: Actual compressed size (0x%x (%u)) != header compressed size (0x%x (%u))!\n",
			(sz_buf - PRELINK_LZSS_FILE_HEADER_SIZE_ON_DISK), (sz_buf - PRELINK_LZSS_FILE_HEADER_SIZE_ON_DISK),
			fhdr.compressed_size, fhdr.compressed_size);

	}

	decomp_buffer_size = (fhdr.uncompressed_size + DECOMP_BUFFER_PAD);

	if (!(p_decomp_buf = (unsigned char *)malloc(decomp_buffer_size))) {
		fprintf(stderr, "Error allocating buffer for decompression!\n");

		return 0;
	}


	printf("Decompressing 0x%x (%u) byte payload ... ", sz_payload, sz_payload); fflush(stdout);

	decomp_ret = decompress_lzss(p_decomp_buf, p_payload, sz_payload);

	printf("Done\n\t(decompressed size = 0x%x (%u))\n", decomp_ret, decomp_ret); fflush(stdout);

	if (decomp_ret != fhdr.uncompressed_size) {
		fprintf(stderr, "WARNING: Actual decompressed size (0x%x (%u)) != header decompressed size (0x%x (%u))!\n",
			decomp_ret, decomp_ret,
			fhdr.uncompressed_size, fhdr.uncompressed_size);
	}

	out_file_size = decomp_ret;

	computed_adler32 = local_adler32(p_decomp_buf, out_file_size);
	if (computed_adler32 != fhdr.adler32) {
		fprintf(stderr, "WARNING: Actual adler32 checksum (0x%x) != header adler32 checksum (0x%x)!\n",
			computed_adler32, fhdr.adler32);
	}


	if (!save_binary_file(out_fname, p_decomp_buf, out_file_size)) {
		fprintf(stderr, "Error writing output file: '%s'!\n", out_fname);
		free((void *)p_decomp_buf);
		return 0;
	}

	printf("Wrote output file: '%s' (size = 0x%x (%u))\n", out_fname, out_file_size, out_file_size); fflush(stdout);

	free((void *)p_decomp_buf);

	return 1;
}

int
encode_lzss_prelink_file(unsigned char *p_buf, unsigned int sz_buf, const char *out_fname)
{
	PRELINK_LZSS_FILE_HEADER	fhdr;
	unsigned char	reserved_2_data[] = PRELINK_LZSS_HDR_RESERVED_DATA_2;
	unsigned char	*p_out_buf, *p_comp_buf;
	unsigned int	out_buffer_size, comp_buffer_size, out_file_size;
	unsigned char	*comp_ret;
	unsigned char	*p;

	memset(&fhdr, 0, sizeof(fhdr));

	fhdr.signature = PRELINK_LZSS_FILE_HEADER__SIGNATURE;
	fhdr.compress_type = PRELINK_LZSS_FILE_HEADER__COMPRESS_TYPE;
	fhdr.uncompressed_size = sz_buf;
	fhdr.adler32 = local_adler32(p_buf, sz_buf);

	strcpy(fhdr.filename, "mach_kernel");
	memcpy(&fhdr.reserved_2, reserved_2_data, sizeof(reserved_2_data));

	// assume compressed size <= uncompressed size
	comp_buffer_size = sz_buf;
	out_buffer_size = (PRELINK_LZSS_FILE_HEADER_SIZE_ON_DISK + comp_buffer_size);

	if (!(p_out_buf = (unsigned char *)malloc(out_buffer_size))) {
		fprintf(stderr, "Error allocating output buffer!\n");
		return 0;
	}

	// zero-out file header space
	memset(p_out_buf, 0, PRELINK_LZSS_FILE_HEADER_SIZE_ON_DISK);

	p_comp_buf = (p_out_buf + PRELINK_LZSS_FILE_HEADER_SIZE_ON_DISK);


	printf("Compressing 0x%x (%u) byte payload ... ", sz_buf, sz_buf); fflush(stdout);


	if (!(comp_ret = compress_lzss(p_comp_buf, comp_buffer_size, p_buf, sz_buf))) {
		fprintf(stderr, "Error compressing payload!\n");
		free((void *)p_comp_buf);
		return 0;
	}
	fhdr.compressed_size = (comp_ret - p_comp_buf);

	out_file_size = (PRELINK_LZSS_FILE_HEADER_SIZE_ON_DISK + fhdr.compressed_size);

	printf("Done\n\t(compressed size = 0x%x (%u))\n", fhdr.compressed_size, fhdr.compressed_size); fflush(stdout);


	p = p_out_buf;
	p = write_mem_BE_U32(fhdr.signature, p);
	p = write_mem_BE_U32(fhdr.compress_type, p);
	p = write_mem_BE_U32(fhdr.adler32, p);
	p = write_mem_BE_U32(fhdr.uncompressed_size, p);
	p = write_mem_BE_U32(fhdr.compressed_size, p);
	p = write_mem_bytes(fhdr.reserved_1, p, sizeof(fhdr.reserved_1));
	p = write_mem_bytes(fhdr.reserved_2, p, sizeof(fhdr.reserved_2));
	p = write_mem_bytes(fhdr.filename, p, sizeof(fhdr.filename));





	if (!save_binary_file(out_fname, p_out_buf, out_file_size)) {
		fprintf(stderr, "Error writing output file: '%s'!\n", out_fname);
		free((void *)p_out_buf);
		return 0;
	}

	printf("Wrote output file: '%s' (size = 0x%x (%u))\n", out_fname, out_file_size, out_file_size); fflush(stdout);


	free((void *)p_out_buf);

	return 1;
}





void
usage(void)
{
	fprintf(stderr, "Usage: %s [-d|-e] in_fname out_fname\n", Pname);
}

